N_("GTK Theme");
N_("Dark Variant");
N_("Cursor Theme");
N_("Cursor Size");
N_("Icon Theme");
N_("Font");
N_("Text Direction");
N_("Left-to-Right");
N_("Right-to-Left");
N_("Window scaling");
N_("Animations");
N_("Rendering Mode");
N_("Similar");
N_("Image");
N_("Recording");
N_("Show Graphic Updates");
N_("Show Baselines");
N_("Show Pixel Cache");
N_("Simulate touchscreen");
N_("GL Rendering");
N_("When needed");
N_("Always");
N_("Disabled");
N_("Software GL");
N_("Software Surfaces");
N_("Texture Rectangle Extension");
