a {
  background-position: top;
}

b {
}

c {
  background-position: left top;
}

d {
  background-position: right top;
}

e {
}

f {
  background-position: top;
}

g {
}

h {
  background-position: left;
}

i {
  background-position: left top;
}

j {
}

k {
}

l {
  background-position: left bottom;
}

m {
  background-position: left;
}

n {
  background-position: left 25%;
}

o {
  background-position: right;
}

p {
  background-position: right top;
}

q {
}

r {
}

s {
  background-position: right bottom;
}

t {
  background-position: right;
}

u {
  background-position: right 25%;
}

v {
  background-position: bottom;
}

w {
}

x {
  background-position: left bottom;
}

y {
  background-position: right bottom;
}

z {
}

ab {
  background-position: bottom;
}

bb {
}

cb {
  background-position: center;
}

db {
  background-position: top;
}

eb {
  background-position: left;
}

fb {
  background-position: right;
}

gb {
  background-position: bottom;
}

hb {
  background-position: center;
}

ib {
  background-position: center 25%;
}

jb {
  background-position: 25%;
}

kb {
  background-position: 25% top;
}

lb {
}

mb {
}

nb {
  background-position: 25% bottom;
}

ob {
  background-position: 25%;
}

pb {
  background-position: 25% 25%;
}
