popover {
}

popover arrow {
  background-color: rgb(0,255,0);
  border-bottom-color: rgb(0,128,0);
  border-bottom-width: 2px;
  border-left-color: rgb(0,128,0);
  border-left-width: 2px;
  border-right-color: rgb(0,128,0);
  border-right-width: 2px;
  border-top-color: rgb(0,128,0);
  border-top-width: 2px;
}

modelbutton:hover.flat {
  background-color: rgb(255,0,0);
}
