a {
  border-image-repeat: initial;
}

b {
  border-image-repeat: inherit;
}

c {
  border-image-repeat: unset;
}

d {
  border-image-repeat: stretch;
}

e {
  border-image-repeat: stretch;
}

f {
  border-image-repeat: repeat;
}

g {
  border-image-repeat: round;
}

h {
  border-image-repeat: space;
}

i {
  border-image-repeat: stretch round;
}

j {
  border-image-repeat: space repeat;
}
