@import "reset-to-defaults.css";

#no-repeat {
  background-image: url("border-image-ball-red.png");
  background-repeat: no-repeat;
}

#repeat {
  background-image: url("border-image-ball-green.png");
  background-repeat: repeat;
}

#default {
  background-image: url("green-20x20.png");
}

#default-ref {
  background-image: none;
  background-color: lime;
}
