a {
  font-family: initial;
}

b {
  font-family: inherit;
}

c {
  font-family: unset;
}

d {
  font-family: "Macaroni";
}

e {
  font-family: "Macaroni";
}

f {
  font-family: "Macaroni", "Spaghetti", "Sans";
}
