a {
  margin-left: calc(1em + 3px);
}

b {
  transition-duration: calc(-200ms + 1s);
}

c {
  margin-left: calc(1em + 2px);
}

d {
  background-size: calc(-10px + 100%);
}

e {
  border-left-width: 2px;
}

f {
  background-size: 24px;
}

g {
  background-size: 1px;
}

h {
  background-size: calc(27em + 36px);
}

i {
  border-top-left-radius: 5px 1px;
}

j {
  border-bottom-right-radius: 5px 1px;
}

k {
  margin-left: 1px;
}
