a {
  margin-bottom: initial;
  margin-left: initial;
  margin-right: initial;
  margin-top: initial;
}

b {
  margin-bottom: inherit;
  margin-left: inherit;
  margin-right: inherit;
  margin-top: inherit;
}

c {
  margin-bottom: unset;
  margin-left: unset;
  margin-right: unset;
  margin-top: unset;
}

d {
  margin-bottom: 0;
  margin-left: 0;
  margin-right: 0;
  margin-top: 0;
}

e {
  margin-bottom: 10px;
  margin-left: 10px;
  margin-right: 10px;
  margin-top: 10px;
}
